import errno
import os


def exist_or_create(folder):
    if not os.path.exists(os.path.dirname(folder)):
        try:
            os.makedirs(os.path.dirname(folder))
        except OSError as exc:
            if exc.errno != errno.EEXIST:
                raise



# -*- coding: utf-8 -*-
import os
import subprocess
from typing import Tuple

import click
from kubernetes import client, config, watch
from kubernetes.client.rest import ApiException
from urllib3.exceptions import MaxRetryError

import src.cli.console as console
from src.local.exceptions import UnikubeClusterUnavailableError


class CMDWrapper(object):
    base_command = None

    class CMDException(Exception):
        pass

    def __init__(self, debug_output=False):
        self._debug_output = debug_output

    def _execute(self, arguments, stdin: str = None) -> subprocess.Popen:
        cmd = [self.base_command] + arguments
        kwargs = self._get_kwargs()
        process = subprocess.Popen(cmd, **kwargs)
        if stdin:
            process.communicate(stdin)
        try:
            process.wait()
        except KeyboardInterrupt:
            try:
                process.terminate()
            except OSError:
                pass
            process.wait()
        return process

    def _run(self, arguments) -> subprocess.CompletedProcess:
        cmd = [self.base_command] + arguments
        return subprocess.run(cmd, env=self._get_environment())

    def _get_kwargs(self):
        kwargs = {"env": self._get_environment(), "encoding": "utf-8"}
        if not self._debug_output:
            kwargs.update(
                {
                    "stdout": subprocess.PIPE,
                    "close_fds": True,
                    "stderr": subprocess.STDOUT,
                }
            )
        return kwargs

    def _get_environment(self):
        env = os.environ.copy()
        return env


class KubeCtl(CMDWrapper):
    base_command = "kubectl"

    def __init__(self, provider_data, debug_output=False):
        if not provider_data.kubeconfig_path:
            raise ValueError("Project does not contain the kubeconfigPath parameter")
        self._provider_data = provider_data
        super(KubeCtl, self).__init__(debug_output)

    def _get_environment(self):
        env = super(KubeCtl, self)._get_environment()
        env["KUBECONFIG"] = self._provider_data.kubeconfig_path
        return env

    def _get_kwargs(self):
        kwargs = super(KubeCtl, self)._get_kwargs()
        kwargs["stdin"] = subprocess.PIPE
        return kwargs

    def create_namespace(self, namespace: str):
        arguments = ["create", "namespace", namespace]
        self._execute(arguments)

    def apply_str(self, namespace, text: str):
        arguments = ["apply", "--namespace", namespace, "-f", "-"]
        self._execute(arguments, text)

    def get_ingress_data(self, namespace):
        arguments = ["get", "ingress", "--namespace", namespace]
        process = self._execute(arguments)
        output = process.stdout.read()
        # skip the header
        ingress_lines = output.split("\n")[1:]
        result = []
        for line in ingress_lines:
            line = list(filter(lambda x: x != "", line.split(" ")))
            if line:
                result.append(
                    {
                        "name": line[0],
                        "hosts": line[1],
                        "address": line[2],
                        "ports": line[3],
                    }
                )
        return result

    def get_pods(self, namespace):
        arguments = ["get", "pods", "--namespace", namespace]
        process = self._execute(arguments)
        output = process.stdout.read()
        return output

    def describe_pod(self, pod, namespace):
        arguments = ["describe", "pod", pod, "--namespace", namespace]
        process = self._execute(arguments)
        output = process.stdout.read()
        return output

    def exec_pod(self, pod, namespace, command, interactive=False):
        arguments = ["exec", "-it", pod]
        if namespace:
            arguments += ["--namespace", namespace]
        arguments += ["--", command]
        if interactive:
            console.info(f"Running '{command}' (interactively) on: {pod}. Please press Ctrl+D to exit.")
            process = self._run(arguments)
            output = process.stdout
        else:
            process = self._execute(arguments)
            output = process.stdout.read()
        return output


class Docker(CMDWrapper):
    base_command = "docker"

    def build(self, tag, context, dockerfile=None, target=None) -> Tuple[bool, str]:
        arguments = ["build", "-t", tag, context]
        if target:
            arguments = arguments + ["--target", target]

        if dockerfile:
            arguments = arguments + ["-f", dockerfile]

        process = self._execute(arguments)
        process = process

        if process.returncode == 0:
            return True, ""
        else:
            return (
                False,
                "Could not build the Docker image, please make sure the image can be built",
            )

    def check_running(self, name):
        arguments = ["ps"]
        process = self._execute(arguments)
        output = process.stdout.read()
        return name in output

    def exec(self, container, command, interactive=False):
        arguments = ["exec", "-it", container, command]
        if interactive:
            console.info(f"Running '{command}' (interactively) on: {container}. Please press Ctrl+D to exit.")
            process = self._run(arguments)
            output = process.stdout
        else:
            process = self._execute(arguments)
            output = process.stdout.read()
        return output


class Telepresence(KubeCtl):
    base_command = "telepresence"

    def swap(
        self,
        deployment,
        image_name,
        command=None,
        namespace=None,
        envs=None,
        mounts=None,
    ):
        arguments = ["--swap-deployment", deployment]
        if namespace:
            arguments = arguments + ["--namespace", namespace]
        arguments = arguments + ["--docker-run", "--rm"]
        if mounts:
            for mount in mounts:
                arguments = arguments + ["-v", f"{mount[0]}:{mount[1]}"]
        if envs:
            for env in envs:
                arguments = arguments + ["--env", f"{env[0]}={env[1]}"]

        # this name to be retrieved for "app shell" command
        arguments = arguments + ["--name", image_name.replace(":", "")]
        arguments.append(image_name)
        if command:
            arguments = arguments + ["sh", "-c"] + [f"{' '.join(command)}"]

        process = self._execute(arguments)
        if process.returncode != 0:
            console.error("There was an error with switching the deployment, please find details above")

    def _get_environment(self):
        env = super(Telepresence, self)._get_environment()
        env["TELEPRESENCE_USE_DEPLOYMENT"] = "1"
        return env


class KubeAPI(object):
    def __init__(self, provider_data, deck):
        self._provider_data = provider_data
        self._deck = deck
        self._namespace = self._deck["namespace"] or "default"
        self._api_client = config.new_client_from_config(provider_data.kubeconfig_path)
        self._core_api = client.CoreV1Api(self._api_client)
        self._networking_api = client.NetworkingV1beta1Api(self._api_client)

    @property
    def is_available(self):
        try:
            self._core_api.get_api_resources()
            return True
        except ApiException:
            return False

    def get_pods(self):
        try:
            return self._core_api.list_namespaced_pod(self._namespace, watch=False)
        except MaxRetryError:
            raise UnikubeClusterUnavailableError

    def get_pod(self, pod_name):
        pods = self.get_pods()
        for pod in pods.items:
            if pod.metadata.name == pod_name:
                return pod
        return None

    def get_logs(self, pod, follow):
        if follow:
            w = watch.Watch()
            try:
                for log in w.stream(
                    self._core_api.read_namespaced_pod_log,
                    name=pod,
                    namespace=self._namespace,
                ):
                    click.echo(log)
            except ApiException as e:
                console.error(str(e))
        else:
            try:
                ret = self._core_api.read_namespaced_pod_log(name=pod, namespace=self._namespace)
            except ApiException as e:
                console.error(str(e))
            else:
                return ret

    def get_ingress(self):
        ret = self._networking_api.list_namespaced_ingress(self._namespace, watch=False)
        return ret

    def get_configmap(self, name=None):
        ret = self._core_api.list_namespaced_config_map(self._namespace, watch=False)
        if name:
            ret = next(filter(lambda x: x.metadata.name == name, ret.items))
        return ret



from abc import ABC, abstractmethod

from src.context.types import ContextData


class UnikubeFileVersionError(Exception):
    pass


class UnikubeFileError(Exception):
    pass


class UnikubeFile(ABC):
    def __init__(self, path: str, data: dict):
        pass

    @abstractmethod
    def get_context(self) -> ContextData:
        raise NotImplementedError



import inquirer

import src.cli.console as console


# input
def list(
    message,
    choices,
    message_no_choices="No choices available!",
):
    # choices exist
    if not len(choices) > 0:
        console.info(message_no_choices)
        return None

    # prompt
    questions = [
        inquirer.List(
            "select",
            message=message,
            choices=choices,
        ),
    ]
    answers = inquirer.prompt(questions)
    if not answers:
        return None

    return answers.get("select", None)



from enum import Enum
from typing import Union

import requests
from gql import Client, gql
from gql.transport.requests import RequestsHTTPTransport
from retrying import retry

import src.cli.console as console
from src import settings


# EnvironmentType
class EnvironmentType(Enum):
    LOCAL = "LOCAL"
    REMOTE = "REMOTE"


class RetryException(Exception):
    pass


# retry exception logic
def retry_exception(exception):
    return isinstance(exception, RetryException)


class GraphQL:
    def __init__(
        self,
        authentication,
        url=settings.GRAPHQL_URL,
        timeout=settings.GRAPHQL_TIMEOUT,
    ):
        self.url = url
        self.timeout = timeout

        # automatic token refresh
        self.authentication = authentication
        self.access_token = str(authentication.general_data.authentication.access_token)

        # client
        self.client = self._client()

    def _client(self):
        # header
        headers = {
            "Content-type": "application/json",
            "Authorization": "Bearer " + str(self.access_token),
        }

        # transport
        transport = RequestsHTTPTransport(
            url=self.url,
            use_json=True,
            headers=headers,
            verify=False,
            retries=3,
            timeout=self.timeout,
        )

        # client
        client = Client(transport=transport)

        return client

    @retry(retry_on_exception=retry_exception, stop_max_attempt_number=2)
    def query(
        self,
        query: str,
        query_variables: dict = None,
    ) -> Union[dict, None]:
        try:
            query = gql(query)
            data = self.client.execute(
                document=query,
                variable_values=query_variables,
            )

        except requests.exceptions.HTTPError:
            # refresh token
            response = self.authentication.refresh()
            if not response["success"]:
                console.exit_login_required()

            self.access_token = response["response"]["access_token"]
            self.client = self._client()

            raise RetryException("retry")

        return data



import click
from requests import HTTPError

import src.cli.console as console
from src.graphql import EnvironmentType, GraphQL
from src.helpers import download_specs
from src.local.system import KubeAPI, KubeCtl
from src.storage.user import get_local_storage_user


@click.command()
@click.option("--organization", "-o", help="Select an organization")
@click.option("--project", "-p", help="Select a project")
@click.pass_obj
def list(ctx, organization=None, project=None, **kwargs):
    """
    List all decks.
    """

    context = ctx.context.get(organization=organization, project=project)

    # GraphQL
    try:
        graph_ql = GraphQL(authentication=ctx.auth)
        data = graph_ql.query(
            """
            query($organization_id: UUID, $project_id: UUID) {
                allDecks(organizationId: $organization_id, projectId: $project_id) {
                    results {
                        id
                        title
                        project {
                            title
                            organization {
                                title
                            }
                        }
                    }
                }
            }
            """,
            query_variables={
                "organization_id": organization,
                "project_id": project,
            },
        )
    except Exception as e:
        data = None
        console.debug(e)
        console.exit_generic_error()

    deck_list = data["allDecks"]["results"]
    if not deck_list:
        console.info("No decks available. Please go to https://app.unikube.io and create a project.")
        exit(0)
    # format list to table
    table_data = []
    for deck in deck_list:
        data = {}

        if not context.organization_id:
            data["organization"] = deck["project"]["organization"]["title"]

        if not context.project_id:
            data["project"] = deck["project"]["title"]

        data["id"] = deck["id"]
        data["title"] = deck["title"]
        table_data.append(data)
    # console
    console.table(data=table_data)


@click.command()
@click.argument("deck_name", required=False)
@click.pass_obj
def info(ctx, deck_name, **kwargs):
    """
    Display further information of the selected deck.
    """

    # GraphQL
    try:
        graph_ql = GraphQL(authentication=ctx.auth)
        data = graph_ql.query(
            """
            {
                allDecks {
                    results {
                        id
                        title
                        description
                        namespace
                        type
                    }
                }
            }
            """
        )
    except Exception:
        data = None
        console.exit_generic_error()

    deck_list = data["allDecks"]["results"]

    # argument
    if not deck_name:
        # argument from context
        context = ctx.context.get()
        if context.deck_id:
            deck = ctx.context.get_deck()
            deck_name = deck["title"]

        # argument from console
        else:
            deck_name = console.list(
                message="Please select a deck",
                choices=[deck["title"] for deck in deck_list],
            )
            if deck_name is None:
                return None

    # select
    deck_selected = None
    for deck in deck_list:
        if deck["title"] == deck_name:
            deck_selected = deck
            break

    # console
    if deck_selected:
        console.table(
            data={
                "key": [k for k in deck_selected.keys()],
                "value": [v for v in deck_selected.values()],
            },
            headers=["Key", "Value"],
        )
    else:
        console.error("Deck does not exist.")


@click.command()
@click.argument("deck_id", required=False)
@click.option("--remove", "-r", is_flag=True, default=False, help="Remove local deck context")
@click.pass_obj
def use(ctx, deck_id, remove, **kwargs):
    """
    Set local deck context.
    """

    # user_data / context
    local_storage_user = get_local_storage_user()
    user_data = local_storage_user.get()
    context = user_data.context

    # option: --remove
    if remove:
        user_data.context.deck_id = None
        local_storage_user.set(user_data)
        console.success("Deck context removed.")
        return None

    # GraphQL
    try:
        graph_ql = GraphQL(authentication=ctx.auth)
        data = graph_ql.query(
            """
            query($organization_id: UUID, $project_id: UUID) {
                allDecks(organizationId: $organization_id, projectId: $project_id) {
                    results {
                        title
                        id
                        project {
                            id
                            organization {
                                id
                            }
                        }
                    }
                }
            }
            """,
            query_variables={
                "organization_id": context.organization_id,
                "project_id": context.project_id,
            },
        )
    except Exception as e:
        data = None
        console.debug(e)
        console.exit_generic_error()

    deck_list = data["allDecks"]["results"]
    deck_dict = {deck["id"]: deck for deck in deck_list}

    # argument
    if not deck_id:
        deck_title = console.list(
            message="Please select a deck",
            choices=[deck["title"] for deck in deck_dict.values()],
        )
        if deck_title is None:
            return False

        for id, deck in deck_dict.items():
            if deck["title"] == deck_title:
                deck_id = id

    deck = deck_dict.get(deck_id, None)
    if not deck:
        console.error(f"Unknown deck with id: {deck_id}.")

    # set deck
    user_data.context.organization_id = deck["project"]["organization"]["id"]
    user_data.context.project_id = deck["project"]["id"]
    user_data.context.deck_id = deck["id"]
    local_storage_user.set(user_data)

    console.success(f"Deck context: {user_data.context}")


@click.command()
@click.argument("deck_title", required=False)
@click.pass_obj
def install(ctx, deck_title, **kwargs):
    """
    Install deck.
    """

    # GraphQL
    try:
        graph_ql = GraphQL(authentication=ctx.auth)
        data = graph_ql.query(
            """
            {
                allDecks(limit:100) {
                    totalCount
                    results {
                        id
                        title
                        namespace
                        environment {
                            id
                            type
                            valuesPath
                        }
                        project {
                            id
                            title
                            organization {
                                title
                            }
                        }
                    }
                }
            }
            """,
            query_variables={},
        )

    except Exception as e:
        data = None
        console.debug(e)
        console.exit_generic_error()

    deck_list = data["allDecks"]["results"]

    # argument
    if not deck_title:
        # argument from context
        context = ctx.context.get()
        if context.deck_id:
            deck = ctx.context.get_deck()
            deck_title = deck["title"]

        # argument from console
        else:
            deck_list_choices = [item["title"] for item in deck_list]
            deck_title = console.list(
                message="Please select a deck",
                choices=deck_list_choices,
            )
            if deck_title is None:
                return False

    # check access to the deck
    deck_title_list = [deck["title"] for deck in deck_list]
    if deck_title not in deck_title_list:
        console.info(f"The deck '{deck_title}' could not be found.")
        return None

    # get project_id
    project_id = None
    deck = None
    for deck in deck_list:
        if deck["title"] == deck_title:
            project_id = deck["project"]["id"]
            break

    # check if cluster is ready
    cluster_data = ctx.cluster_manager.get(id=project_id)
    cluster = ctx.cluster_manager.select(cluster_data=cluster_data)
    if not cluster:
        console.error("The project cluster does not exist.")
        return None

    # check if kubernetes cluster is running/ready
    if not cluster.ready():
        console.info(f"Kubernetes cluster for '{cluster.display_name}' is not running")
        return None

    # check cluster level
    try:
        environment = EnvironmentType(deck["environment"][0]["type"])
    except Exception as e:
        console.debug(e)
        environment = None

    if environment != EnvironmentType.LOCAL:
        console.error("This deck cannot be installed locally.")
        return None

    # download manifest
    try:
        environment_id = deck["environment"][0]["id"]
        general_data = ctx.storage_general.get()
        console.info("Now requesting manifests. This process may takes a few seconds.")
        all_specs = download_specs(
            access_token=general_data.authentication.access_token,
            environment_id=environment_id,
        )
    except HTTPError as e:
        if e.response.status_code == 404:
            console.warning(
                "This deck does potentially not specify a valid Environment of type 'local'. "
                f"Please go to https://app.unikube.io/project/{project_id}/decks "
                f"and save a valid values path."
            )
            exit(1)
        else:
            console.error("Could not load manifest: " + str(e))
            exit(1)

    provider_data = cluster.storage.get()

    # KubeCtl
    kubectl = KubeCtl(provider_data=provider_data)

    namespace = deck["namespace"]
    kubectl.create_namespace(namespace)
    with click.progressbar(
        all_specs,
        label="[Info] Installing Kubernetes resources to the cluster",
    ) as files:
        for file in files:
            kubectl.apply_str(namespace, file["content"])
    console.info("The cluster is currently applying all changes, this may takes several minutes")

    ingresss = KubeAPI(provider_data, deck).get_ingress()
    ingress_data = []
    for ingress in ingresss.items:
        hosts = []
        paths = []
        for rule in ingress.spec.rules:
            hosts.append(f"http://{rule.host}:{provider_data.publisher_port}")  # NOSONAR
            for path in rule.http.paths:
                paths.append(f"{path.path} -> {path.backend.service_name}")
                # this is an empty line in output
            hosts.append("")
            paths.append("")

        ingress_data.append(
            {
                "name": ingress.metadata.name,
                "url": "\n".join(hosts),
                "paths": "\n".join(paths),
            }
        )

    # console
    console.table(
        ingress_data,
        headers={"name": "Name", "url": "URLs"},
    )


@click.command()
@click.argument("deck_name", required=False)
def uninstall(deck_name, **kwargs):
    raise NotImplementedError


@click.command()
@click.argument("deck_name", required=False)
@click.option("--app", "-a", help="Stream aggregated logs")
def logs(deck_name, **kwargs):
    raise NotImplementedError


@click.command()
@click.argument("deck_name", required=False)
@click.option("--app", "-a", help="Request a new environment variable")
def request_env(deck_name, **kwargs):
    raise NotImplementedError
